/*
    NESLA NullLogic Embedded Scripting Language
    Copyright (C) 2007-2009 Dan Cahill

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "nesla/libnesla.h"
#include "nesla/libmath.h"
#include <math.h>
NES_FUNCTION(neslamath_math2)
{
	obj_t *cobj1=nes_getobj(N, &N->l, "1");
	obj_t *cobj2=nes_getobj(N, &N->l, "2");
	char *fname=nes_getstr(N, &N->l, "0");
	num_t n;

	if (cobj1->val->type!=NT_NUMBER) n_error(N, NE_SYNTAX, fname, "expected a number for arg1");
	if (nc_strcmp(fname, "ceil")==0) {
		n=ceil(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "floor")==0) {
		n=floor(cobj1->val->d.num);
	/* add trig functions */
	} else if (nc_strcmp(fname, "acos")==0) {
		n=acos(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "asin")==0) {
		n=asin(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "atan")==0) {
		n=atan(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "atan2")==0) {
		if (cobj2->val->type!=NT_NUMBER) n_error(N, NE_SYNTAX, fname, "expected a number for arg2");
		n=atan2(cobj1->val->d.num, cobj2->val->d.num);
	} else if (nc_strcmp(fname, "cos")==0) {
		n=cos(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "sin")==0) {
		n=sin(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "tan")==0) {
		n=tan(cobj1->val->d.num);
	/* add exp and log functions */
	} else if (nc_strcmp(fname, "exp")==0) {
		n=exp(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "log")==0) {
		if (cobj1->val->d.num<=0) n_error(N, NE_SYNTAX, fname, "arg must not be zero");
		n=log(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "log10")==0) {
		if (cobj1->val->d.num<=0) n_error(N, NE_SYNTAX, fname, "arg must not be zero");
		n=log10(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "pow")==0) {
		if (cobj2->val->type!=NT_NUMBER) n_error(N, NE_SYNTAX, fname, "expected a number for arg2");
		n=pow(cobj1->val->d.num, cobj2->val->d.num);
	/* add hyperbolic functions */
	} else if (nc_strcmp(fname, "cosh")==0) {
		n=cosh(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "sinh")==0) {
		n=sinh(cobj1->val->d.num);
	} else if (nc_strcmp(fname, "tanh")==0) {
		n=tanh(cobj1->val->d.num);
	/* add other functions */
	} else if (nc_strcmp(fname, "sqrt")==0) {
		n=sqrt(cobj1->val->d.num);
	} else {
		n=0;
	}
	nes_setnum(N, &N->r, "", n);
	return 0;
}
int neslamath_register_all(nes_state *N)
{
	obj_t *tobj;

	tobj=nes_settable(N, &N->g, "math");
	tobj->val->attr|=NST_HIDDEN;
	nes_setcfunc(N, tobj, "acos",  (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "asin",  (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "atan",  (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "atan2", (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "cos",   (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "sin",   (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "tan",   (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "exp",   (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "log",   (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "log10", (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "cosh",  (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "sinh",  (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "tanh",  (NES_CFUNC)neslamath_math2);
	nes_setcfunc(N, tobj, "sqrt",  (NES_CFUNC)neslamath_math2);
	return 0;
}

#ifdef PIC
DllExport int neslalib_init(nes_state *N)
{
	neslamath_register_all(N);
	return 0;
}
#endif
